#include <jmp.h>
#include <monitor_dialog.h>
#include <hash.h>
#include <heap_dump.h>
#include <gtk/gtk.h>
#include <jthread.h>

/** Temporary buffer for labeling. */
#define BLEN 32
static char buf[BLEN];

static char* get_monitor_name (monitor_info* mi) {
    return mi->name;
}

static char* get_monitor_id (monitor_info* mi) {
    snprintf (buf, BLEN, "%p", mi->id);
    return buf;    
}

/** return a safe string for the name of the thread. */
static char* get_thread_name (jthread* t) {
    return t ? jthread_get_thread_name (t) : _("<unknown>");
}

static char* get_monitor_owner (monitor_info* mi) {
    return get_thread_name (mi->owner);
}

static char* get_monitor_entry_count (monitor_info* mi) {
    snprintf (buf, BLEN, "%d", mi->entry_count);
    return buf;        
}

static void add_threads_to_frame (int num, jthread** threads, GtkWidget* frametable) {
    GtkWidget* label;
    int j = 0;
    for (j = 0; j < num; j++) {
	/** thread name */
	label = gtk_label_new (get_thread_name (threads[j]));
	gtk_table_attach (GTK_TABLE (frametable), label, 0, 1, j, j + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	
	/** id */
	snprintf (buf, BLEN, "id: %p", jthread_get_env_id (threads[j]));
	label = gtk_label_new (buf);
	gtk_table_attach (GTK_TABLE (frametable), label, 1, 2, j, j + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);

	/** group */
	label = gtk_label_new (jthread_get_group_name (threads[j]));
	gtk_table_attach (GTK_TABLE (frametable), label, 2, 3, j, j + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	
	/** parent */
	label = gtk_label_new (jthread_get_parent_name (threads[j]));
	gtk_table_attach (GTK_TABLE (frametable), label, 3, 4, j, j + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    }
}

static void add_monitor (void* monitor, void* notebook) {
    GtkWidget* label;
    GtkWidget* table;    
    GtkWidget* frame;
    GtkWidget* frametable;
    int i = 0;
    monitor_info* mi = (monitor_info*)monitor;
    GtkWidget* tab = (GtkWidget*)notebook;

    gchar *labels[] = { _("Name"),
			_("Monitor id"),
			_("Owner"), 
			_("Entry count"), 
			NULL};
    
    char* ((*mp[])(monitor_info* m)) = { get_monitor_name,
					 get_monitor_id,
					 get_monitor_owner, 
					 get_monitor_entry_count,
					 NULL};

    table = gtk_table_new (7, 2, FALSE);
    gtk_table_set_row_spacings (GTK_TABLE (table), 2);
    gtk_table_set_col_spacings (GTK_TABLE (table), 2);

    while (labels[i] != NULL) {
	label = gtk_label_new (labels[i]);
	gtk_table_attach (GTK_TABLE (table), label, 0, 1, i, i + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    
	label = gtk_label_new (mp[i](mi));
	gtk_table_attach (GTK_TABLE (table), label, 1, 2, i, i + 1, 
			  GTK_FILL, 0, 0, 0);
	gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT);
	gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
	i++;
    } 

    frame = gtk_frame_new (_("Waiting to enter"));
    frametable = gtk_table_new (mi->num_waiting_to_enter, 4, FALSE);
    gtk_container_add (GTK_CONTAINER (frame), frametable);
    gtk_table_set_row_spacings (GTK_TABLE (frametable), 2);
    gtk_table_set_col_spacings (GTK_TABLE (frametable), 10);
    gtk_table_attach (GTK_TABLE (table), frame,  0, 2, i, i + 1, 
		      GTK_FILL, 0, 0, 0);
    add_threads_to_frame (mi->num_waiting_to_enter, mi->waiting_to_enter, frametable);

    i++;
    frame = gtk_frame_new (_("Waiting for notify"));
    frametable = gtk_table_new (mi->num_waiting_for_notify, 4, FALSE);
    gtk_container_add (GTK_CONTAINER (frame), frametable);
    gtk_table_set_row_spacings (GTK_TABLE (frametable), 2);
    gtk_table_set_col_spacings (GTK_TABLE (frametable), 10);
    gtk_table_attach (GTK_TABLE (table), frame,  0, 2, i, i + 1, 
		      GTK_FILL, 0, 0, 0);
    add_threads_to_frame (mi->num_waiting_for_notify, mi->waiting_for_notify, frametable);

    
    label = gtk_label_new (mi->name);
    gtk_notebook_append_page (GTK_NOTEBOOK (tab), table, label);
}

void show_monitor_dialog (hashtab* monitors) {
    GtkWidget* win;
    GtkWidget* tab;

    win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data (GTK_OBJECT (win), "win", win);
    gtk_window_set_title (GTK_WINDOW (win), _("Monitor info"));
    tab = gtk_notebook_new ();
    gtk_notebook_popup_enable (GTK_NOTEBOOK (tab));
    gtk_notebook_set_scrollable (GTK_NOTEBOOK (tab), TRUE);
    gtk_container_add (GTK_CONTAINER (win), tab);
    
    jmphash_for_each_with_arg ((jmphash_iter_fa)add_monitor, monitors, tab);    
    
    gtk_widget_show_all (win);    
}
